"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const _ = require("lodash");
const types_1 = require("../typings/types");
const api_1 = require("./api");
const github_1 = require("./github");
/**
 * Start the convergence for an entire workspace
 */
function convergeWorkspace(workspaceId, sdm, options) {
    return __awaiter(this, void 0, void 0, function* () {
        // Look for SCMProviders of type github_com
        const graphClient = sdm.configuration.graphql.client.factory.create(workspaceId, sdm.configuration);
        const providers = yield graphClient.query({
            name: "ScmProvider",
            variables: {
                type: options.providerType,
            },
            options: automation_client_1.QueryNoCacheOptions,
        });
        if (providers && providers.SCMProvider && providers.SCMProvider.length > 0) {
            const provider = providers.SCMProvider[0];
            yield convergeProvider(provider, graphClient);
        }
    });
}
exports.convergeWorkspace = convergeWorkspace;
/**
 * Converge a single provider
 *
 * This is main entry point into the GitHub webhook convergence logic
 */
// tslint:disable:cyclomatic-complexity
function convergeProvider(provider, graphClient) {
    return __awaiter(this, void 0, void 0, function* () {
        // Only deal with auth'ed providers
        if (!provider.credential || !provider.credential.secret || !provider.targetConfiguration) {
            return automation_client_1.Success;
        }
        const token = provider.credential.secret;
        const orgs = provider.targetConfiguration.orgSpecs || [];
        const repos = provider.targetConfiguration.repoSpecs || [];
        const errors = [];
        let state;
        for (const org of _.uniq(orgs)) {
            try {
                automation_client_1.logger.info(`Converging GitHub org '${org}'`);
                yield convergeOrg(org, provider, token, graphClient);
            }
            catch (e) {
                automation_client_1.logger.error(`Error converging GitHub org '${org}': `, e);
                if (github_1.isAuthError(e)) {
                    state = types_1.ScmProviderStateName.unauthorized;
                    errors.push(`Authorization error occurred converging GitHub org '${org}'`);
                }
                else {
                    state = state === types_1.ScmProviderStateName.unauthorized ? types_1.ScmProviderStateName.unauthorized : types_1.ScmProviderStateName.misconfigured;
                    errors.push(`Failed to converge GitHub org '${org}': ${github_1.printError(e)}`);
                }
            }
        }
        for (const repo of _.uniqWith(repos, _.isEqual)) {
            const slug = `${repo.ownerSpec}/${repo.nameSpec}`;
            try {
                automation_client_1.logger.info(`Converging GitHub repo '${slug}'`);
                yield convergeRepo(repo.ownerSpec, repo.nameSpec, provider, token, graphClient);
            }
            catch (e) {
                automation_client_1.logger.error(`Error converging GitHub slug '${slug}': `, e);
                if (github_1.isAuthError(e)) {
                    state = types_1.ScmProviderStateName.unauthorized;
                    errors.push(`Authorization error occurred converging GitHub repo '${slug}'`);
                }
                else {
                    state = state === types_1.ScmProviderStateName.unauthorized ? types_1.ScmProviderStateName.unauthorized : types_1.ScmProviderStateName.misconfigured;
                    errors.push(`Failed to converge GitHub repo '${slug}': ${github_1.printError(e)}`);
                }
            }
        }
        const webhooksToDelete = [];
        // Delete webhooks for orgs or repos that went away; for now only mark them to get deleted later
        for (const webhook of (yield api_1.loadProvider(graphClient, provider.id)).webhooks) {
            const org = webhook.tags.find(t => t.name === "org");
            const repo = webhook.tags.find(t => t.name === "repo");
            const hookId = webhook.tags.find(t => t.name === "hook_id");
            if (!!org && !provider.targetConfiguration.orgSpecs.some(o => o === org.value)) {
                automation_client_1.logger.info(`Deleting GitHub webhook on org '${org.value}' because it is no longer in target configuration`);
                webhooksToDelete.push(webhook.id);
                if (!!hookId) {
                    try {
                        yield github_1.gitHub(token, provider).orgs.deleteHook({
                            hook_id: +hookId.value,
                            org: org.value,
                        });
                    }
                    catch (e) {
                        automation_client_1.logger.info(`Failed to delete GitHub webhook on org '${org.value}'`);
                    }
                }
            }
            else if (!!repo) {
                const slug = repo.value.split("/");
                if (!provider.targetConfiguration.repoSpecs.some(o => o.ownerSpec === slug[0] && o.nameSpec === slug[1])) {
                    automation_client_1.logger.info(`Deleting GitHub webhook on repo '${repo.value}' because it is no longer in target configuration`);
                    webhooksToDelete.push(webhook.id);
                    if (!!hookId) {
                        try {
                            yield github_1.gitHub(token, provider).repos.deleteHook({
                                hook_id: +hookId.value,
                                owner: slug[0],
                                repo: slug[1],
                            });
                        }
                        catch (e) {
                            automation_client_1.logger.info(`Failed to delete GitHub webhook on repo '${repo.value}'`);
                        }
                    }
                }
            }
        }
        // Mark all hooks with no hook_id to get deleted
        for (const webhook of (yield api_1.loadProvider(graphClient, provider.id)).webhooks) {
            const hookId = webhook.tags.find(t => t.name === "hook_id");
            if (!hookId) {
                automation_client_1.logger.info(`Deleting webhook because of missing hook_id`);
                webhooksToDelete.push(webhook.id);
            }
        }
        // Finally delete webhooks that got marked for deletion
        for (const webhookToDelete of _.uniq(webhooksToDelete)) {
            yield api_1.deleteWebhook(graphClient, webhookToDelete);
        }
        yield api_1.setProviderState(graphClient, provider, state, errors);
        return automation_client_1.Success;
    });
}
exports.convergeProvider = convergeProvider;
// tslint:enable:cyclomatic-complexity
/**
 * Converge a single org
 */
function convergeOrg(org, provider, token, graphClient) {
    return __awaiter(this, void 0, void 0, function* () {
        let createWebbook = false;
        const webhook = (provider.webhooks || []).find(wh => (wh.tags || []).some(t => t.name === "org" && t.value === org));
        if (!webhook) {
            // Case 1: no webhook on provider for given org
            automation_client_1.logger.info(`No webhook found for GitHub org '${org}'. Creating new webhook`);
            createWebbook = true;
        }
        else {
            const tag = webhook.tags.find(wh => wh.name === "hook_id");
            if (tag) {
                const hookId = tag.value;
                try {
                    const githubWebhook = yield github_1.gitHub(token, provider).orgs.getHook({
                        hook_id: +hookId,
                        org,
                    });
                    if (!githubWebhook.data.active || githubWebhook.data.config.url !== webhook.url) {
                        // Case 2: webhook on provider different to the one on GitHub
                        automation_client_1.logger.info(`Webhook found for GitHub org '${org}' on SCM provider but webhook is inactive or different url. ` +
                            `Deleting and creating new webhook`);
                        yield api_1.deleteWebhook(graphClient, webhook.id);
                        yield github_1.gitHub(token, provider).orgs.deleteHook({
                            hook_id: +hookId,
                            org,
                        });
                        createWebbook = true;
                    }
                }
                catch (e) {
                    // Case 3: webhook on provider but non on GitHub
                    automation_client_1.logger.info(`Webhook found for GitHub org '${org}' on SCM provider but webhook not found on GitHub. Deleting and creating new webhook`);
                    yield api_1.deleteWebhook(graphClient, webhook.id);
                    createWebbook = true;
                }
            }
            else {
                automation_client_1.logger.info(`Webhook found for GitHub org '${org}' on SCM provider but no hook_id. Deleting and creating new webhook`);
                yield api_1.deleteWebhook(graphClient, webhook.id);
                createWebbook = true;
            }
        }
        if (createWebbook) {
            yield github_1.createOrgWebhook(org, provider, token, graphClient);
        }
        // Link up org and owner
        yield graphClient.mutate({
            name: "SetOwnerLogin",
            variables: {
                providerId: provider.providerId,
                owner: org,
                login: provider.credential.owner.person.scmId.login,
            },
        });
    });
}
exports.convergeOrg = convergeOrg;
/**
 * Converge a single repo
 */
// tslint:disable:cyclomatic-complexity
function convergeRepo(owner, repo, provider, token, graphClient) {
    return __awaiter(this, void 0, void 0, function* () {
        let createWebbook = false;
        const slug = `${owner}/${repo}`;
        const webhook = (provider.webhooks || []).find(wh => (wh.tags || []).some(t => t.name === "repo" && t.value === slug));
        // Check if repo is covered by an org webhook
        const orgHasHook = (provider.targetConfiguration.orgSpecs || []).some(o => o === owner);
        if (!webhook) {
            // Case 1: no webhook on provider for given org
            automation_client_1.logger.info(`No webhook found for GitHub repo '${slug}'. Creating new webhook`);
            createWebbook = true;
        }
        else {
            const tag = webhook.tags.find(wh => wh.name === "hook_id");
            if (tag) {
                const hookId = tag.value;
                try {
                    const githubWebhook = yield github_1.gitHub(token, provider).repos.getHook({
                        hook_id: +hookId,
                        owner,
                        repo,
                    });
                    if (orgHasHook || !githubWebhook.data.active || githubWebhook.data.config.url !== webhook.url) {
                        // Case 2: webhook on provider different to the one on GitHub
                        automation_client_1.logger.info(`Webhook found for GitHub repo '${slug}' on SCM provider but webhook is inactive or different url or the org has a hook. ` +
                            `Deleting and creating new webhook`);
                        yield api_1.deleteWebhook(graphClient, webhook.id);
                        yield github_1.gitHub(token, provider).repos.deleteHook({
                            hook_id: +hookId,
                            owner,
                            repo,
                        });
                        createWebbook = true;
                    }
                }
                catch (e) {
                    // Case 3: webhook on provider but non on GitHub
                    automation_client_1.logger.info(`Webhook found for GitHub repo '${slug}' on SCM provider but webhook not found on GitHub. Deleting and creating new webhook`);
                    yield api_1.deleteWebhook(graphClient, webhook.id);
                    createWebbook = true;
                }
            }
            else {
                automation_client_1.logger.info(`Webhook found for GitHub repo '${repo}' on SCM provider but no hook_id. Deleting and creating new webhook`);
                yield api_1.deleteWebhook(graphClient, webhook.id);
                createWebbook = true;
            }
        }
        if (orgHasHook) {
            return;
        }
        if (createWebbook) {
            yield github_1.createRepoWebhook(owner, repo, provider, token, graphClient);
        }
        // Link up repo and owner
        yield graphClient.mutate({
            name: "SetRepoLogin",
            variables: {
                providerId: provider.providerId,
                owner,
                repo,
                login: provider.credential.owner.person.scmId.login,
            },
        });
    });
}
exports.convergeRepo = convergeRepo;
// tslint:enable:cyclomatic-complexity
//# sourceMappingURL=converge.js.map